# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

import os,stat

try:
    import pip

    # install uvicorn if not present (for DevApp mode)
    try:
        import uvicorn
    except ImportError as e:
        pip.main(['install', 'uvicorn[standard]'])

    # install starlette if not present (for DevApp mode)
    try:
        import starlette
    except ImportError as e:
        pip.main(['install', 'starlette'])

    devappmode = True
except:
    devappmode = False


code = """
# -*- coding: utf-8 -*-
# the simplest htag'app, in the best env to start development (hot reload/refresh)

from htag import Tag

class App(Tag.body):
    statics="body {background:#EEE;}"

    def init(self):
        self += "Hello World"

#=================================================================================
%s
# from htag.runners import BrowserHTTP as Runner
# from htag.runners import ChromeApp as Runner

app=Runner(App)
if __name__=="__main__":
    app.run()
""" % (devappmode and "from htag.runners import DevApp as Runner" or "from htag.runners import BrowserHTTP as Runner")

newfile = "main.py"

if not os.path.isfile(newfile):
    with open(newfile,"w+") as fid:
        fid.write(code)

    print("HTag App file created --> main.py")
else:
    print(f"It seems that you've already got a {newfile} file")



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################


from .. import Tag
from ..render import HRenderer
from . import common

import os

import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse,JSONResponse
from starlette.routing import Route

#="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="
# mainly code from the good old guy ;-)
#="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="
import sys
import shutil
import tempfile
import subprocess
import logging
import threading
import webbrowser

logger = logging.getLogger(__name__)

class FULLSCREEN: pass
CHROMECACHE=".cache"

class _ChromeApp:
    def __init__(self, url, appname="driver",size=None,lockPort=None,chromeargs=[]):

        def find_chrome_win():
            import winreg  # TODO: pip3 install winreg

            reg_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe"
            for install_type in winreg.HKEY_CURRENT_USER, winreg.HKEY_LOCAL_MACHINE:
                try:
                    with winreg.OpenKey(install_type, reg_path, 0, winreg.KEY_READ) as reg_key:
                        return winreg.QueryValue(reg_key, None)
                except WindowsError:
                    pass

        def find_chrome_mac():
            default_dir = "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"
            if os.path.exists(default_dir):
                return default_dir


        if sys.platform[:3] == "win":
            exe = find_chrome_win()
        elif sys.platform == "darwin":
            exe = find_chrome_mac()
        else:
            for i in ["chromium-browser", "chromium", "google-chrome", "chrome"]:
                try:
                    exe = webbrowser.get(i).name
                    break
                except webbrowser.Error:
                    exe = None

        if not exe:
            raise Exception("no chrome browser, no app-mode !")
        else:
            args = [ #https://peter.sh/experiments/chromium-command-line-switches/
                exe,
                "--app=" + url, # need to be a real http page !
                "--app-id=%s" % (appname),
                "--app-auto-launched",
                "--no-first-run",
                "--no-default-browser-check",
                "--disable-notifications",
                "--disable-features=TranslateUI",
                #~ "--no-proxy-server",
            ] + chromeargs
            if size:
                if size == FULLSCREEN:
                    args.append("--start-fullscreen")
                else:
                    args.append( "--window-size=%s,%s" % (size[0],size[1]) )

            if lockPort: #enable reusable cache folder (coz only one instance can be runned)
                self.cacheFolderToRemove=None
                args.append("--remote-debugging-port=%s" % lockPort)
                args.append("--disk-cache-dir=%s" % CHROMECACHE)
                args.append("--user-data-dir=%s/%s" % (CHROMECACHE,appname))
            else:
                self.cacheFolderToRemove=os.path.join(tempfile.gettempdir(),appname+"_"+str(os.getpid()))
                args.append("--user-data-dir=" + self.cacheFolderToRemove)
                args.append("--aggressive-cache-discard")
                args.append("--disable-cache")
                args.append("--disable-application-cache")
                args.append("--disable-offline-load-stale-cache")
                args.append("--disk-cache-size=0")

            logger.debug("CHROME APP-MODE: %s"," ".join(args))
            # self._p = subprocess.Popen(args)
            self._p = subprocess.Popen(args,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            #~ if lockPort:
                #~ http_client = tornado.httpclient.HTTPClient()
                #~ self._ws = None
                #~ while self._ws == None:
                    #~ try:
                        #~ url = http_client.fetch("http://localhost:%s/json" % debugport).body
                        #~ self._ws = json.loads(url)[0]["webSocketDebuggerUrl"]
                    #~ except Exception as e:
                        #~ self._ws = None

    def wait(self):
        self._p.wait()

    def __del__(self): # really important !
        self._p.kill()
        if self.cacheFolderToRemove: shutil.rmtree(self.cacheFolderToRemove, ignore_errors=True)

    #~ def _com(self, payload: dict):
        #~ """ https://chromedevtools.github.io/devtools-protocol/tot/Browser/#method-close """
        #~ payload["id"] = 1
        #~ r=json.loads(wsquery(self._ws,json.dumps(payload)))["result"]
        #~ return r or True

    #~ def focus(self): # not used
        #~ return self._com(dict(method="Page.bringToFront"))

    #~ def navigate(self, url): # not used
        #~ return self._com(dict(method="Page.navigate", params={"url": url}))

    def exit(self):
        #~ self._com(dict(method="Browser.close"))
        self._p.kill()
#="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="="


class ChromeApp:
    """ This a "Chrome App Runner" : it runs the front in a "Chrome App mode" by re-using
    the current chrome installation, in a headless mode.
    (it's a little bit like pywebview/cef technics, but based on the real installed chrome)

    BTW : it uses Starlette/http as backend server
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.hrenderer = None
        self.tagClass = tagClass

    def _run_the_base(self,port=8707):
        host = "127.0.0.1"
        self._chromeapp = _ChromeApp(f"http://{host}:{port}",size=(800,600))

        asgi=Starlette(debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            Route('/', self.POST, methods=["POST"]),
        ])

        self._server = threading.Thread(name='ChromeAppServer', target=uvicorn.run,args=(asgi,),kwargs=dict(host=host, port=port))


    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.hrenderer and self.hrenderer.init == init:
            return self.hrenderer

        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""
        return HRenderer(self.tagClass, js, self._chromeapp.exit, init=init )


    async def GET(self,request) -> HTMLResponse:
        self.hrenderer = self.instanciate( str(request.url) )
        return HTMLResponse( str(self.hrenderer) )

    async def POST(self,request) -> JSONResponse:
        data = await request.json()
        dico = await self.hrenderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event"))
        return JSONResponse(dico)

    def run(self,port=8707):
        self._run_the_base(port)
        self._server.start()
        self._chromeapp.wait()
        os._exit(0) # to force quit the thread/uvicorn server



# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common


import os,json
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route,WebSocketRoute
from starlette.endpoints import WebSocketEndpoint

class BrowserStarletteWS(Starlette):
    """ Simple ASync Web Server (with starlette) with WebSocket interactions with HTag.
        Open the rendering in a browser tab.

        The instance is an ASGI htag app
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)
        self.hrenderer = None
        self.tagClass = tagClass

        class WsInteract(WebSocketEndpoint):
            encoding = "json"

            async def on_receive(this, websocket, data):
                actions = await self.hrenderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event"))
                await websocket.send_text( json.dumps(actions) )

        Starlette.__init__(self,debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            WebSocketRoute("/ws", WsInteract),
        ])


    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.hrenderer and self.hrenderer.init == init:
            return self.hrenderer

        js = """
async function interact( o ) {
    ws.send( JSON.stringify(o) );
}

var ws = new WebSocket("ws://"+document.location.host+"/ws");
ws.onopen = start;
ws.onmessage = function(e) {
    action( JSON.parse(e.data) );
};
"""
        return HRenderer(self.tagClass, js, lambda: os._exit(0), init=init)

    async def GET(self,request):

        self.hrenderer=self.instanciate( str(request.url) )

        return HTMLResponse( str(self.hrenderer) )

    def run(self, host="127.0.0.1", port=8000, openBrowser=True):   # localhost, by default !!
        import uvicorn,webbrowser
        if openBrowser:
            webbrowser.open_new_tab(f"http://{host}:{port}")

        uvicorn.run(self, host=host, port=port)


# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common



""" REAL WEB http,
- .exit() has no effect ;-)
- can handle multiple client (with htuid cookie)
- can handle multiple "tag class", and brings query_params as "query_params:dict" when instanciate tag
- the session are purged automatically after timeout/5m of inactivity
- "http get query_params" are passed at Tag'init, if it accepts a query_params:dict param. (needed for https://htag.glitch.me)
"""

import uuid
import time
import asyncio
import logging

from starlette.applications import Starlette
from starlette.responses import HTMLResponse,JSONResponse,Response
from starlette.routing import Route
logger = logging.getLogger(__name__)

class WebHTTP(Starlette):
    """ Simple ASync Web Server (with starlette) with HTTP interactions with htag.
        can handle multiple instances & multiples Tag

        The instance is an ASGI htag app
    """
    def __init__(self,tagClass:type, timeout=5*60):
        assert issubclass(tagClass,Tag)
        self.sessions={} # nb htag instances (htag x user)
        self.tagClass=tagClass
        self.timeout=timeout

        Starlette.__init__(self,debug=True, routes=[
            Route('/',   self.GET,   methods=["GET"]),
            Route('/{sesid}',   self.POST,  methods=["POST"]),
        ], on_startup=[self._purgeSessions])

    async def _purgeSessions(self):

        async def purge():
            while True:
                now=time.time()
                for ses_id,ses_info in list(self.sessions.items()):
                    if now - ses_info["lastaccess"] > self.timeout:
                        logger.info("PURGE SESSION: %s (sessions:%s)", ses_id, len(self.sessions))
                        del self.sessions[ses_id]

                await asyncio.sleep(60) # check every 60s

        asyncio.ensure_future( purge() )

    async def GET(self,request) -> HTMLResponse:
        print( "NB SES=",len(self.sessions) )
        return self.serve(request, self.tagClass )

    def serve(self,request, klass, init=None) -> HTMLResponse:
        """ Serve for the `request`, an instance of the class 'klass'
        initialized with `init` (tuple (*args,**kargs))
        if init is None : takes them from request.url ;-)

        return an htmlresponse (htag init page to start all)
        """
        htuid = request.cookies.get('htuid') or str(uuid.uuid4())

        if init is None:
            # no init params
            # so we take thoses from the url
            init = common.url2ak( str(request.url) )
        else:
            assert type(init)==tuple
            assert type(init[0])==tuple
            assert type(init[1])==dict

        hr = self.instanciate(htuid, klass, init )

        r = HTMLResponse( str(hr) )
        r.set_cookie("htuid",htuid,path="/")
        return r

    def instanciate(self, htuid, klass, init) -> HRenderer:
        """ get|create an instance of `klass` for user session `htuid`
        (get|save it into self.sessions)
        """
        sesid = f"{klass.__name__}|{htuid}"   # there can be only one instance of klass, at a time !

        if sesid in self.sessions and self.sessions[sesid]["renderer"].init == init:
            # same url (same klass/params), same htuid -> same instance
            hr=self.sessions[sesid]["renderer"]
        else:
            # url has changed ... recreate an instance
            js = """
                async function interact( o ) {
                    action( await (await window.fetch("/%s",{method:"POST", body:JSON.stringify(o)})).json() )
                }

                window.addEventListener('DOMContentLoaded', start );
            """ % sesid

            hr=HRenderer(klass, js, init=init ) # NO EXIT !!

        # update session info
        self.sessions[ sesid ] = dict(
            lastaccess=time.time(),
            renderer=hr,
        )

        return hr


    async def POST(self,request) -> Response:
        sesid=request.path_params.get('sesid',None)

        if sesid in self.sessions:
            hr = self.sessions[ sesid ]["renderer"]

            logger.info("INTERACT WITH SESSION %s (sessions:%s)",sesid,len(self.sessions))
            data=await request.json()
            actions = await hr.interact(data["id"],data["method"],data["args"],data["kargs"])
            return JSONResponse(actions)

        else:
            # session expired or bad call
            return HTMLResponse( "400 BAD REQUEST" , status_code=400 )



    def run(self, host="0.0.0.0", port=8000):   # wide, by default !!
        import uvicorn
        uvicorn.run(self, host=host, port=port)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common

import os

from starlette.applications import Starlette
from starlette.responses import HTMLResponse,JSONResponse
from starlette.routing import Route


class BrowserStarletteHTTP(Starlette):
    """ Simple ASync Web Server (with starlette) with HTTP interactions with htag.
        Open the rendering in a browser tab.

        The instance is an ASGI htag app
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.hrenderer = None
        self.tagClass = tagClass

        Starlette.__init__(self,debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            Route('/', self.POST, methods=["POST"]),
        ])

    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.hrenderer and self.hrenderer.init == init:
            return self.hrenderer

        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""

        return HRenderer(self.tagClass, js, lambda: os._exit(0), init=init)

    async def GET(self,request) -> HTMLResponse:
        self.hrenderer = self.instanciate( str(request.url) )
        return HTMLResponse( str(self.hrenderer) )

    async def POST(self,request) -> JSONResponse:
        data = await request.json()
        dico = await self.hrenderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event"))
        return JSONResponse(dico)

    def run(self, host="127.0.0.1", port=8000, openBrowser=True):   # localhost, by default !!
        import uvicorn,webbrowser
        if openBrowser:
            webbrowser.open_new_tab(f"http://{host}:{port}")

        uvicorn.run(self, host=host, port=port)




# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common

import os,json,sys,asyncio
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route,WebSocketRoute
from starlette.endpoints import WebSocketEndpoint

import socket

def isFree(ip, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(1)
    return not (s.connect_ex((ip,port)) == 0)

class DevApp(Starlette):
    """ DEV APP, Runner specialized for development process. Features :
            * autoreload on file changes
            * refresh UI/HTML/client part, after server autoreloaded
            * console.log/info in devtools, for all exchanges
            * uvicorn debug
            * js error() method auto implemented (popup with skip/refresh)

        Simple ASync Web Server (with starlette) with WebSocket interactions with HTag.
        Open the rendering in a browser tab.

        The instance is an ASGI htag app
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.hrenderers={}
        self.tagClass=tagClass

        class WsInteract(WebSocketEndpoint):
            encoding = "json"

            async def on_receive(this, websocket, data):
                className = data["class"]
                actions = await self.hrenderers[className].interact(data["id"],data["method"],data["args"],data["kargs"],data["event"])
                await websocket.send_text( json.dumps(actions) )

        Starlette.__init__(self,debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            WebSocketRoute("/ws", WsInteract),
        ])


    def instanciate(self, tagClass, init):
        """ intanciate or reuse """
        className = tagClass.__name__

        if className in self.hrenderers and self.hrenderers[className].init == init:
            # the current version has already be initialized, we return the saved instance
            return self.hrenderers[className]

        #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
        # add a Static Template, for displaying beautiful full error on UI ;-)
        #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\ #TODO: perhaps something integrated in hrenderer
        styles={"color":"yellow","text-decoration":"none"}
        t=Tag.H.div( _style="z-index:10000000000;position:fixed;top:10px;left:10px;background:#F00;padding:8px;border:1px solid yellow" )
        t <= Tag.H.a("X",_href="#",_onclick="this.parentNode.remove()",_style=styles,_title="Forget error (skip)")
        t <= " "
        t <= Tag.H.a("REFRESH",_href="#",_onclick="window.location.reload()",_style=styles,_title="Restart the UI part by refreshing it")
        t <= Tag.H.pre(_style="overflow:auto")
        template = Tag.H.template(t,_id="DevAppError")
        #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

        js = """

window.error=function(txt) {
    var clone = document.importNode(document.querySelector("#DevAppError").content, true);
    clone.querySelector("pre").innerHTML = txt
    document.body.appendChild(clone)
}

async function interact( o ) {
    o["class"] = "%s";
    let packet = JSON.stringify(o)
    console.info("[htag interact]",packet.length,o)
    ws.send( packet );
}

var ws = new WebSocket("ws://"+document.location.host+"/ws");
ws.onopen = function() {console.info("[htag start]");start()};
ws.onclose = function() {document.body.innerHTML="Refreshing";window.location.reload()}

ws.onmessage = function(e) {
    let data = JSON.parse(e.data);
    console.info("[htag action]",e.data.length,data)
    action( data );
};
""" % className
        self.hrenderers[className]=HRenderer(tagClass, js, self.killme, fullerror=True, statics=[template,], init=init)
        return self.hrenderers[className]


    async def GET(self,request):
        return self.serve(request, self.tagClass )

    def serve(self,request, klass, init=None) -> HTMLResponse:
        """ Serve for the `request`, an instance of the class 'klass'
        initialized with `init` (tuple (*args,**kargs))
        if init is None : takes them from request.url ;-)

        return an htmlresponse (htag init page to start all)
        """
        if init is None:
            # no init params
            # so we take thoses from the url
            init = common.url2ak( str(request.url) )
        else:
            assert type(init)==tuple
            assert type(init[0])==tuple
            assert type(init[1])==dict

        hr = self.instanciate( klass, init )

        return HTMLResponse( str(hr) )


    def run(self, host="127.0.0.1", port=8000, openBrowser=True):   # localhost, by default !!
        import uvicorn,webbrowser
        import inspect,sys
        from pathlib import Path

        try:
            fi= inspect.getframeinfo(sys._getframe(1))
            stem = Path(fi.filename).stem
            instanceName = fi.code_context[0].strip().split(".")[0]
        except Exception as e:
            print("Can't run DevApp :",e)
            sys.exit(-1)

        fileapp = stem+":"+instanceName
        url = f"http://{host}:{port}"
        print("="*79)
        print(f"Start Uvicorn Reloader for '{fileapp}' ({url})")
        print("="*79)

        os.environ["PYTHONUNBUFFERED"]="FALSE"

        if openBrowser:
            webbrowser.open_new_tab(url)

        uvicorn.run(fileapp,host=host,port=port,reload=True,debug=True)

        # config = uvicorn.Config(self,host=host,port=port,reload=True,debug=True)
        # server = uvicorn.Server(config=config)
        # loop = asyncio.get_event_loop()
        # loop.run_until_complete(server.run())
        # server.run()

    def killme(self):
        #TODO: exit() should work on devapp
        os._exit(0) # but not, coz uvicorn restart the dead process in reloader mode




# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################
import json,asyncio,types,traceback

from . import __version__
from .tag import HTagException,H, Tag, TagBase, BaseCaller

from typing import Callable, Optional

StrNonable = Optional[str]


import logging
logger = logging.getLogger(__name__)

def fmtcaller(method:str,a,k) -> str:
    """ pretty format caller method( a:tuple, k:dict) -> str"""
    MAX=15
    def show(x):
        if isinstance(x,str):
            if len(x)>=MAX:
                x=x[:MAX-3]+"..."
            return json.dumps(x)
        elif isinstance(x,bytes):
            if len(x)>=MAX:
                x= x[:MAX-3]+b"..."
        return str(x)

    args = [show(i) for i in a]+[f"{k}={show(v)}" for k,v in k.items()]
    return f'{method}( %s )' % ", ".join(args)

class Stater:
    """ Class to save all states of 'tag' and children/embbeded Tags, before interaction!
        so, after interaction : .guess() can guess modified one
        (intelligent rendering)
    """
    def __init__(self, tag: Tag ):
        logger.debug("Stater.__init__(), save state before interactions ")
        self.tag=tag
        self._states={}

        def rec( childs ):
            for obj in childs:
                tag,childs= list(obj.items())[0]
                self._states[id(tag)] = tag._getStateImage()
                rec( childs )

        rec( [self.tag._getTree()])
        logger.debug("Stater.__init__(), save %s states", len(self._states))

    def guess(self): # -> list<Tag>
        """ to be runned after interactions to guess whose are modifieds
            return modifieds tags
        """
        logger.debug("Stater.guess(), start guessing ....")
        modifieds=[]

        def rec( childs):
            for obj in childs:
                tag,childs= list(obj.items())[0]
                state_after = tag._getStateImage()
                state_before= self._states.get( id(tag) )
                if state_after != state_before and isinstance(tag,Tag):
                    logger.debug("STATE BEFORE for %s = '%s'", repr(tag), state_before )
                    logger.debug("STATE AFTER  for %s = '%s'", repr(tag), state_after )
                    modifieds.append(tag)
                else:
                    # no need to see childs, coz the parent will redraw all (childs too)
                    rec(childs)

        rec( [self.tag._getTree()] )
        logger.debug("Stater.guess(), modifieds components : %s", [repr(i) for i in modifieds])
        return modifieds


class HRenderer:
    def __init__(self, tagClass: type, js:str, exit_callback:Optional[Callable]=None, init= ((),{}), fullerror=False, statics=[] ):
        """ tag object will be setted as a 'body' tag ! """
        """ js should containt an interact method, and an initializer to call the start()"""
        if not issubclass(tagClass, Tag): raise HTagException("HRenderer can only handle tag subclasses !")
        if not isinstance(statics, list): raise HTagException("HRenderer statics should be a list !")
        self.fullerror = fullerror
        self._interaction_scripts=[]
        self.init = tuple( [tuple(init[0]),dict(init[1])] ) # save args/kargs whose initialized the instance

        try:
            args,kargs = init
            kargs["hr"] = self
            tag = tagClass( *args,**kargs )
        except TypeError:
            logger.warning(f"Can't instanciate tag '{tagClass.__name__}' with {init} arguments, so instanciate it without argument !")
            kargs={"hr": self}
            tag = tagClass(**kargs)

        self.tag=tag
        self.tag.tag="body" # force first tag as body !!
        if exit_callback:
            # add an .exit() method on the tag !!!!!!!!!!!!!!!!!!!!!!
            self.tag.exit = exit_callback

        self._statics: list = list(statics)

        ensureList=lambda x: list(x) if type(x) in [list,tuple] else [x]

        def feedStatics(tag):
            for i in ensureList(tag.statics):
                if isinstance(i,TagBase):
                    if isinstance(i,Tag):
                        i = i._ensureTagBase()

                    if i.md5 not in [j.md5 for j in self._statics]:
                        self._statics.append( i )
                elif isinstance(i,str): # auto add as Tag.style // CSS
                    self._statics.append( Tag.H.style(i))
                elif isinstance(i,bytes): # auto add as Tag.script // JS
                    self._statics.append( Tag.H.script(i.decode()))
                else:
                    raise HTagException("Included static is bad")


        if hasattr(self.tag, "imports") and self.tag.imports is not None:
            # there is an "imports" attribut
            # so, try to import statics according "imports" attrs on tags
            logger.info("Include statics from Tag's imports attibut")
            feedStatics(self.tag)
            def rec( tag ):
                if hasattr(tag, "imports") and tag.imports is not None:
                    imports = ensureList(tag.imports)
                    if not all([isinstance(c,type) and issubclass(c,TagBase) for c in imports]):
                        raise HTagException("imports can contains only Tag classes")
                    for c in imports:
                        feedStatics(c)
                        rec(c)

            rec( self.tag )
        else:
            # there is no "imports" attribut
            # so try to imports statics using Tag subclasses
            logger.info("Include statics from Tag's subclasses")
            def rec( tag ):
                for c in tag.__subclasses__():
                    feedStatics(c)
                    rec(c)

            rec(Tag)

        logger.debug("HRenderer(), statics found : %s", [repr(i) for i in self._statics])

        js_base="""
function start() { %s }

function _error(txt,env) {
    if(window.error)
        error( env+" ERROR: "+txt );
    else
        console.log( env+" ERROR:", txt );
}

function action( o ) {

    if(o.hasOwnProperty("update"))
        Object.keys(o["update"]).forEach(key => {
            document.getElementById( key ).outerHTML = o["update"][key];
        });
    if(o.hasOwnProperty("stream"))
        Object.keys(o["stream"]).forEach(key => {
            document.getElementById( key ).innerHTML += o["stream"][key];
        });

    if(o.hasOwnProperty("post")) {try{ eval( o["post"] )} catch(e) {_error(e, "JS");throw e}};
    if(o.hasOwnProperty("next")) {try{ eval( o["next"] )} catch(e) {_error(e, "JS");throw e}};
    if(o.hasOwnProperty("err")) _error( o["err"], "PYTHON")
}

function jevent (e) {
    let n={};
    if(e===undefined || e==null) return n;

    // pointerevent
    n.isTrusted=e.isTrusted;
    n.altKey=e.altKey;
    n.altitudeAngle=e.altitudeAngle;
    n.azimuthAngle=e.azimuthAngle;
    n.bubbles=e.bubbles;
    n.button=e.button;
    n.buttons=e.buttons;
    n.cancelBubble=e.cancelBubble;
    n.cancelable=e.cancelable;
    n.clientX=e.clientX;
    n.clientY=e.clientY;
    n.composed=e.composed;
    n.ctrlKey=e.ctrlKey;
    n.defaultPrevented=e.defaultPrevented;
    n.detail=e.detail;
    n.eventPhase=e.eventPhase;
    n.height=e.height;
    n.isPrimary=e.isPrimary;
    n.layerX=e.layerX;
    n.layerY=e.layerY;
    n.metaKey=e.metaKey;
    n.movementX=e.movementX;
    n.movementY=e.movementY;
    n.offsetX=e.offsetX;
    n.offsetY=e.offsetY;
    n.pageX=e.pageX;
    n.pageY=e.pageY;
    n.pointerId=e.pointerId;
    n.pointerType=e.pointerType;
    n.pressure=e.pressure;
    n.returnValue=e.returnValue;
    n.screenX=e.screenX;
    n.screenY=e.screenY;
    n.shiftKey=e.shiftKey;
    n.tangentialPressure=e.tangentialPressure;
    n.tiltX=e.tiltX;
    n.tiltY=e.tiltY;
    n.timeStamp=e.timeStamp;
    n.twist=e.twist;
    n.type=e.type;
    n.which=e.which;
    n.width=e.width;
    n.x=e.x;
    n.y=e.y;

    // keyboardevent specific
    n.charCode=e.charCode;
    n.code=e.code;
    n.isComposing=e.isComposing;
    n.key=e.key;
    n.keyCode=e.keyCode;
    n.location=e.location;
    n.repeat=e.repeat;

    return n;
}

%s
""" % (BaseCaller( None ), js)

        self._statics.append( H.script( js_base ))

        self._loop={}   # for savng generator, to keep them during GC


    def _addInteractionScript(self, js:str):
        self._interaction_scripts.append(js)

    async def interact(self,oid,method_name:str,args,kargs,event=None) -> dict:
        """ call the 'method_name' of pyobj 'id' with (args,kargs), return the pyobj/tag"""
        try:
            ## self._interaction_scripts=[] # reset the list
            next_js_call=None

            if oid==0:
                # start (not a real interaction, just produce the first rendering of the main tag (will be a body))
                logger.info("INTERACT INITIAL: %s",repr(self.tag))
                rep = self._mkReponse( [self.tag] )
                rep["update"]={0: list(rep["update"].values())[0]}  #INPERSONNATE (for first interact on id#0)
            else:

                state = Stater(self.tag)
                obj = Tag.find_tag(oid)
                if obj:
                    obj.event=event or {}

                    # call the method
                    method=getattr(obj,method_name)

                    logger.info(f"INTERACT with METHOD {fmtcaller(method_name,args,kargs)}, of %s", repr(obj) )

                    if asyncio.iscoroutinefunction( method ):
                        r=await method(*args,**kargs)
                    else:
                        r=method(*args,**kargs)

                    if isinstance(r, types.AsyncGeneratorType) or isinstance(r, types.GeneratorType):
                        # save it, to avoid GC
                        self._loop[id(r)] = dict(gen=r,tag=obj)

                else:
                    obj = self._loop.get(oid,None)
                    if obj: # it's a existing generator
                        r, obj = obj["gen"], obj["tag"]
                        logger.info("INTERACT with GENERATOR %s(...), of %s",r.__name__, repr(obj) )
                    else:
                        raise Exception(f"{oid} is not an existing Tag or generator (dead objects ?)?!")

                ret=None
                if isinstance(r, types.AsyncGeneratorType):
                    # it's a "async def yield"
                    try:
                        ret = await r.__anext__()
                        next_js_call = BaseCaller(r)
                    except StopAsyncIteration:
                        del self._loop[ id(r) ]
                elif isinstance(r, types.GeneratorType):
                    # it's a "def yield"
                    try:
                        ret = r.__next__()
                        next_js_call = BaseCaller(r)
                    except StopIteration:
                        del self._loop[ id(r) ]
                else:
                    # it's a simple method
                    assert r is None
                    ret=None

                rep= self._mkReponse(state.guess() )

                if ret:
                    obj.add(ret) # add content in object
                    if not isinstance(ret,str) and hasattr(ret,"__iter__"):
                        rep.setdefault("stream",{})[ id(obj) ] = "".join([str(i) for i in ret])
                    else:
                        rep.setdefault("stream",{})[ id(obj) ] = str(ret)


            # --> rep
            if self._interaction_scripts:
                # some js scripts was generated during interactions
                if "post" not in rep:
                    rep["post"]=""
                # add them
                rep["post"] += "\n".join(self._interaction_scripts)

                self._interaction_scripts=[]    # reset the list !

            if next_js_call:
                # if there was generator, set the next js call !
                rep["next"]= str(next_js_call)

        except Exception as e:
            print("ERROR",e)
            fullerror=traceback.format_exc()
            logger.error("Exception %s:\n%s", e, fullerror)
            rep={"err": fullerror if self.fullerror else str(e) }

        logger.info("RETURN --> %s",json.dumps(rep,indent=4))

        return rep

    def _mkReponse(self, tags) -> dict: # can't set a "script" key -> so named "post" (coz guy)
        rep={}

        scripts = []
        updates={}

        if tags:
            logger.debug("Force Tag rendering (for response): %s",[repr(i) for i in tags])
            for tag in tags:
                updates[id(tag)]=str(tag)

                if isinstance(tag,Tag):
                    scripts.extend( tag._getAllJs() )

        if updates:
            rep["update"]=updates

        if scripts:
            rep["post"]="\n".join( scripts )

        return rep

    def __str__(self) -> str:
        head=H.head()
        head <= H.meta(_charset="utf-8")
        head <= H.meta(_name="viewport",_content="width=device-width, initial-scale=1")
        head <= H.meta(_name="version",_content=f"htag {__version__}")
        head <= self._statics
        head <= H.title( self.title )   # set a default title

        body=H.body( "Loading...", _id=0 ) # IMPERSONATE (first interact on id #0)
        return "<!DOCTYPE html>"+str(H.html( head+body ))

    @property
    def title(self) -> str:
        return self.tag.__class__.__name__





# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common

import asyncio
import socket
import webbrowser,os,json
import urllib.parse

class BrowserHTTP:
    """ Simple ASync Web Server with HTTP interactions with HTag. (only stdlib!)
        Open the rendering in a browser tab.

        Should not be used AS IS in a real app ...
        But it's the perfect runner, to test/debug, coz interactions are easier !
    """

    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.hrenderer=None
        self.tagClass=tagClass

    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.hrenderer and self.hrenderer.init == init:
            return self.hrenderer

        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""
        return HRenderer(self.tagClass, js, lambda: os._exit(0), init=init)

    def run(self, host="127.0.0.1", port=8000, openBrowser=True ):   # localhost, by default !!
        """
        ASyncio http server with stdlib ;-)
        Inspired from https://www.pythonsheets.com/notes/python-asyncio.html
        """

        def make_header(type="text/html"):
            header  =  "HTTP/1.1 200 OK\r\n"
            header += f"Content-Type: {type}\r\n"
            header +=  "\r\n"
            return header

        async def handler(conn):

            CHUNK_LIMIT=256 # must be minimal
            req = b''
            while True:
                chunk = await loop.sock_recv(conn, CHUNK_LIMIT)
                if chunk:
                    req += chunk
                    if len(chunk) < CHUNK_LIMIT:
                        break
                else:
                    break


            try:
                # if req.startswith(b"GET / HTTP"):
                if req.startswith(b"GET /"):
                    url=req.decode()[4:].split(" HTTP")[0]

                    info = urllib.parse.urlsplit(url)
                    if info.path=="/":
                        self.hrenderer = self.instanciate(url)
                        resp = make_header()
                        resp += str(self.hrenderer)
                    else:
                        resp = "HTTP/1.1 404 NOT FOUND\r\n"
                elif req.startswith(b"POST / HTTP"):
                    _,content = req.split(b"\r\n\r\n")
                    data = json.loads(content.decode())
                    dico = await self.hrenderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event") )
                    resp = make_header("application/json")
                    resp += json.dumps(dico)
                else:
                    resp = "HTTP/1.1 404 NOT FOUND\r\n"
            except Exception as e:
                print("SERVER ERROR:",e)
                resp = "HTTP/1.1 500 SERVER ERROR\r\n"

            await loop.sock_sendall(conn, resp.encode())
            conn.close()

        async def server(sock, loop):
            while True:
                conn, addr = await loop.sock_accept(sock)
                loop.create_task(handler(conn))

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.setblocking(False)
            s.bind((host, port))
            s.listen(10)

            loop = asyncio.get_event_loop()
            try:
                if openBrowser:
                    webbrowser.open_new_tab(f"http://{host}:{port}")
                loop.run_until_complete(server(s, loop))
            except KeyboardInterrupt:
                pass
            finally:
                loop.close()



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common

import os,json,sys,asyncio

import tornado.ioloop
import tornado.web


class BrowserTornadoHTTP:
    """ Simple ASync Web Server (with TORNADO) with HTTP interactions with htag.
        Open the rendering in a browser tab.
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.hrenderer=None
        self.tagClass=tagClass

        try: # https://bugs.python.org/issue37373 FIX: tornado/py3.8 on windows
            if sys.platform == 'win32':
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        except:
            pass

    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.hrenderer and self.hrenderer.init == init:
            return self.hrenderer

        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""
        return HRenderer(self.tagClass, js, lambda: os._exit(0), init=init)


    def run(self, host="127.0.0.1", port=8000, openBrowser=True):   # localhost, by default !!

        class MainHandler(tornado.web.RequestHandler):
            async def get(this):
                self.hrenderer = self.instanciate( str(this.request.uri) )
                this.write( str(self.hrenderer) )
            async def post(this):
                data = json.loads( this.request.body.decode() )
                dico = await self.hrenderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event"))
                this.write(json.dumps(dico))

        if openBrowser:
            import webbrowser
            webbrowser.open_new_tab(f"http://{host}:{port}")

        app = tornado.web.Application([(r"/", MainHandler),])
        app.listen(port)
        tornado.ioloop.IOLoop.current().start()



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer


import asyncio,os
import webview


"""
LIMITATION :

til pywebview doen't support async jsapi ...(https://github.com/r0x0r/pywebview/issues/867)
it can't work for "async generator" with the asyncio.run() trick (line 50)

pywebview doesn't support :
 - document location change
 - handling query_params from url
"""
#window.addEventListener('pywebviewready', async function() { await window.pywebview.api.start() });

class PyWebView:
    """ Open the rendering in a pywebview instance
        Interactions with builtin pywebview.api ;-)
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        js = """
async function interact( o ) {
    action( await pywebview.api.interact( o["id"], o["method"], o["args"], o["kargs"], o["event"] ) );
}

window.addEventListener('pywebviewready', start );
"""

        self.renderer=HRenderer(tagClass, js, lambda: os._exit(0))

    def run(self):
        class Api:
            def interact(this,tagid,method,args,kargs,event):
                return asyncio.run(self.renderer.interact(tagid,method,args,kargs,event))

        window = webview.create_window(self.renderer.title, html=str(self.renderer), js_api=Api(),text_select=True)
        webview.start(debug=False)

    def exit(self,rc=0):
        os._exit(rc)



# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

class _RunnerNotFunctionnal:
    def __init__(self,*tagClass:type):
        raise ERROR


# mono instance
from .browserhttp import BrowserHTTP

# multi instances
from .pyscript import PyScript


try:
    # mono instance
    from .browserstarlettehttp import BrowserStarletteHTTP
except Exception as err:
    ERROR=err
    class BrowserStarletteHTTP(_RunnerNotFunctionnal): pass

try:
    # mono instance
    from .browserstarlettews import BrowserStarletteWS
except Exception as err:
    ERROR=err
    class BrowserStarletteWS(_RunnerNotFunctionnal): pass

try:
    # mono instance
    from .devapp import DevApp                                      # special dev
except Exception as err:
    ERROR=err
    class DevApp(_RunnerNotFunctionnal): pass

try:
    # multi instance
    from .webhttp import WebHTTP
except Exception as err:
    ERROR=err
    class WebHTTP(_RunnerNotFunctionnal): pass

try:
    # mono instance (just one pywebview ;-)
    from .pywebview import PyWebView
    from .pywebview import PyWebView as PyWebWiew
except Exception as err:
    ERROR=err
    class PyWebWiew(_RunnerNotFunctionnal): pass

try:
    # mono instance (just one chrome app instance)
    from .chromeapp import ChromeApp
except Exception as err:
    ERROR=err
    class ChromeApp(_RunnerNotFunctionnal): pass

try:
    # mono instance (just one android app instance)
    from .androidapp import AndroidApp
except Exception as err:
    ERROR=err
    class AndroidApp(_RunnerNotFunctionnal): pass

try:
    # mono instance
    from .browsertornadohttp import BrowserTornadoHTTP
except Exception as err:
    ERROR=err
    class BrowserTornadoHTTP(_RunnerNotFunctionnal): pass






class StrClass:
    """ mimic the js tag.classList """
    def __init__(self,txt=None):
        if txt is None:
            self._ll=[]
        else:
            if isinstance(txt,list):
                self._ll = list(txt)
            else:
                self._ll=[i.strip() for i in str(txt).split(" ") if i.strip()]

    def contains(self,c) -> int:
        return self._ll.count(c)

    def __contains__(self,kk):
        return self.contains(kk)>0

    def add(self,*cc):
        for c in cc:
            if not self.contains(c):
                self._ll.append(c)

    def remove(self,*cc):
        for c in cc:
            if self.contains(c):
                self._ll.remove(c)

    def toggle(self,c):
        if c in self._ll:
            self.remove(c)
        else:
            self.add(c)

    @property
    def list(self)-> list:
        """ return the python's list, to access more python list methods ;-) """
        return self._ll

    def __add__(self,t:str):
        return StrClass(str(self) + t)
    def __radd__(self,t:str):
        return StrClass(t + str(self))

    def __eq__(self,x):
        return str(self)==str(x)

    def __len__(self):
        return len(self._ll)

    def __str__(self):
        return " ".join(self._ll)
    def __repr__(self):
        return " ".join(self._ll)

#---------------------------------------------------------------

keyize = lambda x: x.strip().lower()


class DictStyle(dict):
    def __init__(self,parent):
        self._parent = parent
        dict.__init__(self,self._parent._ll)

    def __setitem__(self, k, v) -> None:
        self._parent.set(k,v,True)

    def clear(self) -> None:
        self._parent._ll=[]

    def update(self, d:dict):
        for k,v in d.items():
            self[k]=v

class StrStyle:
    """ expose basic methods set/get/contains/remove """
    def __init__(self,txt=None):
        self._ll=[]
        if txt is not None:
            if isinstance(txt,dict):
                self._ll = list(txt.items())
            else:
                for i in str(txt).split(";"):
                    if i and ":" in i:
                        k,v=i.strip().split(":",1)
                        self._ll.append( (keyize(k),v.strip()) )
    @property
    def dict(self):
        """ return a python's dict, to access more python dict methods ;-) """
        return DictStyle(self)

    def set(self,k,v,unique=False):
        if unique: self.remove(k)
        self._ll.append( (keyize(k),v.strip()) )

    def get(self,kk) -> list:
        return [v for k,v in self._ll if keyize(kk)==k]

    def contains(self,kk) -> int:
        return len(self.get(kk))

    def __contains__(self,kk):
        return self.contains(kk)>0

    def remove(self,kk) -> bool:
        ll=[(k,v) for k,v in self._ll if keyize(kk)==k]
        for i in ll:
            self._ll.remove(i)
        return len(self._ll)>0

    def __add__(self,t:str):
        assert ":" in t
        return StrStyle(str(self) + t)
    def __radd__(self,  t:str):
        assert ":" in t
        return StrStyle(t + str(self))

    def __eq__(self,x):
        return str(self)==str(x)

    def __len__(self):
        return len(self._ll)

    def __str__(self):
        return "".join(["%s:%s;" %(k,v) for k,v in self._ll])
    def __repr__(self):
        return "".join(["%s:%s;" %(k,v) for k,v in self._ll])




# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################


from htag import Tag
from htag.render import HRenderer
from . import common

import os,json,asyncio

import tornado.ioloop
import tornado.web
import tornado.platform.asyncio

from threading import Thread

class WebServer(Thread): # the webserver is ran on a separated thread

    def __init__(self,instance,port):
        super(WebServer, self).__init__()
        self.instance=instance
        self.port=port

    def run(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        tornado.platform.asyncio.AsyncIOMainLoop().install()

        class MainHandler(tornado.web.RequestHandler):
            async def get(this):
                self.instance.instanciate( str(this.request.uri)  )
                this.write( str(self.instance.renderer) )
            async def post(this):
                data = json.loads( this.request.body.decode() )
                dico = await self.instance.renderer.interact(data["id"],data["method"],data["args"],data["kargs"],data.get("event"))
                this.write(json.dumps(dico))

        app = tornado.web.Application([(r"/", MainHandler),])
        app.listen(self.port)

        self.loop=asyncio.get_event_loop()
        self.loop.run_forever()


class AndroidApp:
    """
    An "Android Runner", for an HTag App. Which will only work on android/platform

    BTW : it uses tornado/http
    """
    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)

        self.renderer=None
        self.tagClass=tagClass

    def instanciate(self,url:str):
        init = common.url2ak(url)
        if self.renderer and self.renderer.init == init:
            return self.renderer

        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""
        self._exiter=None
        hr=HRenderer(self.tagClass, js, self.go_exit, init=init)
        self.renderer=hr
        return hr

    def go_exit(self):
        if self._exiter is None:
            os._exit(0)
        else:
            self._exiter()

    def run(self): # basically, the same code as guy.runAndroid()
        host,port= "127.0.0.1", 12458
        urlStartPage = f"http://{host}:{port}"

        self.server = WebServer(self,port)

        #=================================================================
        import kivy
        from kivy.app import App
        from kivy.utils import platform
        from kivy.uix.widget import Widget
        from kivy.clock import Clock
        from kivy.logger import Logger

        def run_on_ui_thread(arg):
            pass

        webView       = None
        webViewClient = None
        activity      = None
        if platform == 'android':
            from jnius import autoclass
            from android.runnable import run_on_ui_thread
            webView       = autoclass('android.webkit.WebView')
            webViewClient = autoclass('android.webkit.WebViewClient')
            activity      = autoclass('org.kivy.android.PythonActivity').mActivity
        else:
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("SHOULD BE RUN ON ANDROID (you are on '%s') fallback to browser !!!" % platform)
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            import webbrowser
            webbrowser.open_new_tab(urlStartPage)
            self.server.start()
            return

        # better: https://karobben.github.io/2021/04/30/Python/kivy_android_webviewer/ ?

        class Wv(Widget):
            def __init__(self, runner ):
                self.f2 = self.create_webview               # ! important
                super(Wv, self).__init__()
                self.visible = False

                def exit_app(*a,**k):
                    activity.finish()
                    App.get_running_app().stop()
                    os._exit(0)

                runner._exiter=exit_app

                runner.server.start()                       # ! important

                Clock.schedule_once(self.create_webview, 0)

            @run_on_ui_thread
            def create_webview(self, *args):
                webview = webView(activity)
                webview.getSettings().setJavaScriptEnabled(True)
                webview.getSettings().setDomStorageEnabled(True)
                webview.setWebViewClient(webViewClient())
                activity.setContentView(webview)
                webview.loadUrl(urlStartPage)               # !important

        class ServiceApp(App):
            def build(this):
                return Wv( self )

        ServiceApp().run()
        #=================================================================




# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer
from . import common

import json

class PyScript:

    def __init__(self,tagClass:type):
        assert issubclass(tagClass,Tag)
        self.tagClass=tagClass

    def run(self,window): # window: "pyscript js.window"

        js = """
interact=async function(o) {
 let actions = await window.interactions( JSON.stringify(o) );
 action( JSON.parse(actions) )
}"""
        self.hr = HRenderer(self.tagClass, js, init=common.url2ak( window.document.location.href ))

        window.interactions = self.interactions
        assert window.document.head, "No <head> in <html>"
        assert window.document.body, "No <body> in <html>"

        # install statics in headers
        window.document.head.innerHTML=""
        for s in self.hr._statics:
            tag=window.document.createElement(s.tag)
            tag.innerHTML = "".join([str(i) for i in s.childs if i is not None])
            for key,value in s.attrs.items():
                setattr(tag, key, value)
            window.document.head.appendChild(tag)

        # install the first object in body
        window.document.body.outerHTML=str(Tag.H.body( _id=0 ))

        # and start the process
        window.start()

    async def interactions(self, o):
        data=json.loads(o)
        actions = await self.hr.interact( data["id"], data["method"], data["args"], data["kargs"], data.get("event") )
        return json.dumps(actions)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['htag', 'htag.runners']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'htag',
    'version': '0.8.4',
    'description': 'GUI toolkit for building GUI toolkits (and create beautiful applications for mobile, web, and desktop from a single python3 codebase)',
    'long_description': '# HTag : "H(tml)Tag"\n\n<img src="https://manatlan.github.io/htag/htag.png" width="100" height="100">\n\n[![Test](https://github.com/manatlan/htag/actions/workflows/unittests.yml/badge.svg)](https://github.com/manatlan/htag/actions/workflows/unittests.yml)\n\n<a href="https://pypi.org/project/htag/">\n    <img src="https://badge.fury.io/py/htag.svg" alt="Package version">\n</a>\n\n\nA new python library to create UI (or UI toolkit), which render nativly in anything which can render **html/js/css**.\nThoses can be a browser, a pywebview, an android/apk, or anything based on cef, depending on an [htag runner](https://manatlan.github.io/htag/runners/)\xa0!\nAs it\'s based on html/js rendering: you can easily mix powerful JS libs with powerful PY3 libs : and make powerful python apps !\n\n * For a **desktop app** : You can use the [PyWebView runner](https://manatlan.github.io/htag/runners/#pywebwiew), which will run the UI in a pywebview container (or "ChromeApp runner", in a local chrome app mode).\xa0\n * For a **web app** : You can use the [WebHTTP runner](https://manatlan.github.io/htag/runners/#webhttp), which will run the UI in a web server, and serve the UI on client side, in a browser.\xa0\n * For a **android app** : You can use the [AndroidApp runner](https://manatlan.github.io/htag/runners/#androidapp), which will run the UI in a kiwi webview thru tornado webserver, and can be embedded in an apk ([recipes](https://github.com/manatlan/htagapk))\n * For a **pyscript app** : you can use the [PyScript runner](https://manatlan.github.io/htag/runners/#pyscript), which will run completly in client side\n\nBut yes … the promise is here : **it\'s a GUI toolkit for building "beautiful" applications for mobile, web, and desktop from a single codebase**.\n\n[DOCUMENTATION](https://manatlan.github.io/htag/)\n\n[DEMO/TUTORIAL](https://htag.glitch.me/)\n\n[Changelog](https://github.com/manatlan/htag/blob/main/changelog.md)\n\n[Available on pypi.org](https://pypi.org/project/htag/)\n\n[Announcement on reddit (22/07/14)](https://www.reddit.com/r/Python/comments/vysnci/htag_a_new_gui_tookit_for_webdesktopandroid_from/)\n\n## To have a look\n\nSee the [demo source code](https://github.com/manatlan/htag/blob/main/examples/demo.py)\n\nTo try it :\n\n    $ pip3 install htag pywebview\n    $ wget https://raw.githubusercontent.com/manatlan/htag/main/examples/demo.py\n    $ python3 demo.py\n\n\n## ROADMAP to 1.0.0\n\n * rock solid (need more tests)\n * setup minimal docs ;-)\n * ~~top level api could change (Tag() -> create a Tag, Tag.mytag() -> create a TagBase ... can be a little bit ambiguous)~~\n * ~~manage "query params" from url to initialize Tags/routes~~\n * ~~I don\'t really like the current way to generate js in interaction : need to found something more solid.~~\n * ~~the current way to initiate the statics is odd (only on real (embedded) Tag\'s) : should find a better way (static like gtag ?!)~~\n\n\n## History\n\nAt the beginning, there was [guy](https://github.com/manatlan/guy), which was/is the same concept as [python-eel](https://github.com/ChrisKnott/Eel), but more advanced.\nOne day, I\'ve discovered [remi](https://github.com/rawpython/remi), and asked my self, if it could be done in a *guy way*. The POC was very good, so I released\na version of it, named [gtag](https://github.com/manatlan/gtag). It worked well despite some drawbacks, but was too difficult to maintain. So I decided to rewrite all\nfrom scratch, while staying away from *guy* (to separate, *rendering* and *runners*)... and [htag](https://github.com/manatlan/htag) was born. The codebase is very short, concepts are better implemented, and it\'s very easy to maintain.\n\n\n',
    'author': 'manatlan',
    'author_email': 'manatlan@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/manatlan/htag',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




import urllib.parse

def url2ak(url:str):
    """ transform the querystring of 'url' to (*args,**kargs)"""
    info = urllib.parse.urlsplit(url)
    args=[]
    kargs={}
    if info.query:
        items=info.query.split("&")
        first=lambda x: x[0]
        for i in items:
            if i:
                if "=" in i:
                    if i.endswith("="):
                        kargs[ i.split("=")[0] ] = None
                    else:
                        tt=list(urllib.parse.parse_qs(i).items())[0]
                        kargs[ tt[0] ] = first(tt[1])
                else:
                    tt=list(urllib.parse.parse_qs("x="+i).items())[0]
                    args.append( first(tt[1]) )
    return tuple(args), dict(kargs)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################
import html,json,hashlib
import logging,types,asyncio
import weakref
from typing import Sequence,Union,Optional,Any,Callable,Type
from .attrs import StrClass,StrStyle

AnyTags = Union[ Optional[Any], Sequence[Any]]
StrNonable = Optional[str]

logger = logging.getLogger(__name__)

class HTagException(Exception): pass

md5= lambda x: hashlib.md5(x.encode('utf-8')).hexdigest()

def stringify(obj):
    def my(obj):
        if isinstance(obj,bytes): # trick to convert b'' -> pure javascript
            return "<:<:%s:>:>" % obj.decode()
    return json.dumps(obj, default=my).replace('"<:<:',"").replace(':>:>"',"")


class Elements(list):
    def __add__(self,  elt: AnyTags):
        if elt is None:
            pass
        elif not isinstance(elt,str) and hasattr(elt,"__iter__"):
            self.extend( list(elt) )
        else:
            self.append(elt)
        return self
    def __radd__(self,  elt: AnyTags):
        if elt is None:
            pass
        elif not isinstance(elt,str) and hasattr(elt,"__iter__"):
            self[:0] = list(elt)
        else:
            self.insert(0,elt)
        return self
    def __str__(self):
        return "".join([str(i) for i in self])

    def __repr__(self):
        return f"<Elements {list(self)}>"


class TagBase:
    """ This is a class helper to produce a "HTML TAG" """
    tag: str="div" # default one

    def __init__(self, content:AnyTags=None,**_attrs):
        self.set(content)

        self._attrs={}
        for k,v in _attrs.items():
            if not k.startswith("_"):
                raise HTagException(f"Can't set attributs without underscore ('{k}' should be '_{k}')") # for convention only ;-(
            else:
                self[ k[1:].replace("_","-") ] = v

        # compute a md5 (to indentify state for statics only now)
        # WARN : attrs or content change -> doesn't affect md5 !
        self.md5 = md5( str(self._attrs) + str(self._childs))

    def __le__(self, elt: AnyTags ):
        self.add(elt)
        return elt

    def __iadd__(self,  elt: AnyTags):
        ''' use "+=" instead of "<=" '''
        self.add(elt)
        return self

    def __add__(self,  elt):
        return Elements([self]) + elt
    def __radd__(self,  elt):
        return elt + Elements([self])

    def clear(self):
        self._childs=Elements()

    def set(self,elt:AnyTags):
        """ a bit like .innerHTML setting (avoid clear+add)"""
        self.clear()
        self.add( elt)

    def add(self,elt:AnyTags):
        """ add an object or a list/tuple of objects """
        self._childs.__add__(elt)

    def remove(self, elt=None):
        """Remove an object(elt) from its childs, or itself (if none), if attached (has parent)"""
        if elt is None:
            if self.parent:
                return self.parent.remove(self)
        else:
            if elt in self._childs:
                self._childs.remove(elt)
                return True


    @property
    def childs(self) -> list:
        return self._childs

    @property
    def attrs(self) -> dict:
        return self._attrs

    def __setitem__(self,attr:str,value):
        # if not isinstance(self,Tag):  #TODO: in the future ;-)
        #     raise HTagException("Can't assign a callback on a tagbase")
        attr=attr.strip().lower()
        if attr == "style":
            self._attrs["style"]=StrStyle(value)
        elif attr.strip().lower() == "class":
            self._attrs["class"]=StrClass(value)
        else:
            self._attrs[attr]=value

    def __getitem__(self,attr:str) -> Any:
        attr=attr.strip().lower()
        r=self._attrs.get(attr,None)
        if r is None:
            if attr == "style":
                self._attrs["style"]=StrStyle()
                return self._attrs["style"]
            elif attr == "class":
                self._attrs["class"]=StrClass()
                return self._attrs["class"]
            else:
                return r
        else:
            return r

    def __str__(self):
        return self.__render__( list(self._attrs.items()) )

    def __render__(self, attrs:list ) -> str:
        rattrs=[]
        for k,v in attrs:
            if v is not None:
                if isinstance(v,bool):
                    if v == True:
                        rattrs.append(k)
                else:
                    if v!="":
                        rattrs.append( '%s="%s"' % (k,html.escape( str(v) )) )
        return """<%(tag)s%(attrs)s>%(content)s</%(tag)s>""" % dict(
            tag=self.tag.replace("_","-"),
            attrs=" ".join([""]+rattrs) if rattrs else "",
            content="".join([str(i) for i in self._childs if i is not None]),   # keeps @id
        )

    @property
    def innerHTML(self) -> str:
        """ return innerHTML of the tagbase, but without @id !!!"""
        def noid(v):
            if isinstance(v,TagBase):
                return str(v._ensureTagBase())
            else:
                return str(v)

        return "".join([noid(i) for i in self._childs if i is not None])        # remove @id


    def _getStateImage(self) -> str: #TODO: could disapear (can make something more inteligent here!)
        """Return a str'image (state) of the object, for quick detection (see Stater())"""

        logger.debug("Force Tag rendering (for state image): %s",repr(self))
        str(self) # force a re-rendering (for builded lately)

        image=lambda x: "[%s]"%id(x) if isinstance(x,Tag) else str(x)
        return """%s%s:%s""" % (
            self.tag,
            self._attrs,
            [image(i) for i in self._childs],
        )

    def _getTree(self) -> dict:
        """ return a tree of TagBase childs """
        ll=[]

        for i in self._childs:
            if isinstance(i,TagBase):
                ll.append( i._getTree() )

        return {self:ll}

    def __repr__(self):
        return f"<{self.__class__.__name__}'{self.tag} {self._attrs.get('id')} (childs:{len(self._childs)})>"

    def _ensureTagBase(self): # TODO: can do a lot better here !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        """ ensure that the tag/tagbase will be rendered as a real tagbase (and childs too) """
        if isinstance(self,Tag): # it REMOVES @ID !!!
            attrs = {"_"+k:v for k,v in self.attrs.items() if k != "id" }
        else:
            attrs = {"_"+k:v for k,v in self.attrs.items()}
        contents = [(i._ensureTagBase() if isinstance(i,TagBase) else i) for i in self.childs]
        t=TagBase( contents, **attrs )
        t.tag = self.tag
        return t



class TagBaseCreator(type):
    def __getattr__(self,name:str) -> Type:
        return type('TagBase%s' % name.capitalize(), (TagBase,), {**TagBase.__dict__,"tag":name})

class H(metaclass=TagBaseCreator): # Html
    def __init__(self):
        raise HTagException("no!")



class TagCreator(type):
    def __getattr__(self,name:str) -> Type:
        if name == "H":
            return H
        else:
            return type('Tag%s' % name.capitalize(), (Tag,), {**Tag.__dict__,"tag":name})

class NotBindedCaller:
    def __init__(self,instance,on_event):
        self.__event = on_event
        self.__instance=instance
        self._befores=[]
        self._afters=[]

    def __add__(self,js:str): # -> X Caller
        if not isinstance(js,str):
            raise HTagException("Can't add non string to a Caller")
        self._afters.append(js)
        return self

    def __radd__(self,js:str): # -> X Caller
        if not isinstance(js,str):
            raise HTagException("Can't radd non string to a Caller")
        self._befores.append(js)
        return self

    # def add(self,js):
    #     self.__add__(js)
    #     return self

    def bind(self,callback:Callable,*args,**kargs): #-> Caller
        c = Caller(self.__instance,callback,args,kargs)
        c.assignEventOnTag( self, self.__event  )
        self.__instance[self.__event]=c
        return c

    def __str__(self):
        return ";".join(self._befores+self._afters)


class Caller(NotBindedCaller):
    def __init__(self,instance, callback:Callable, args:tuple, kargs:dict):
        super().__init__(instance,None)
        if not callable(callback): raise HTagException("The caller must be callable !")
        self.instance = instance
        self.callback = callback
        self.args = args
        self.kargs = kargs

        self._others=[]
        self._assigned = None

    # def prior(self,callback:Callable,*args,**kargs):
    #     """ make the current (those params) first, and move the previous in _others
    #         (only the instance keep the same !)
    #     """
    #     self._others.append( (self.callback,self.args,self.kargs) )
    #     self.callback = callback
    #     self.args = args
    #     self.kargs = kargs
    #     return self

    def bind(self,callback:Callable,*args,**kargs): # -> Caller
        if callback is None : # do nothing
            return self
        else:
            if not callable(callback): raise HTagException("The caller must be callable !")
            if any( [isinstance(i,bytes) for i in args] ) or any( [isinstance(i,bytes) for k,i in kargs.items()] ):
                raise HTagException("Can't bind (bytes)'js_arg' in next binders !")
            self._others.append( (callback,args,kargs) )
            return self

    def assignEventOnTag(self, object, event:str):
        self._assigned = "%s-%s" % (event,id(object)) # unique identifier for the event 'event' of the tag 'object'
        self.instance._callbacks_[self._assigned]=self   # save the Caller in _callbacks_ of self.instance
        return self

    def __str__(self) -> str:
        if not self._assigned:
            raise HTagException("Caller can't be serizalized, it's not _assign'ed to an event !")
        newargs = tuple([self._assigned]+list(self.args))
        bc=BaseCaller(self.instance,"__on__", newargs, self.kargs)
        bc._befores = self._befores
        bc._afters = self._afters
        return str(bc)

class BaseCaller(NotBindedCaller):
    def __init__(self,instance, mname:StrNonable=None, args=None, kargs=None):
        super().__init__(instance, None)

        self.instance = instance
        self.mname = mname
        self.args = args
        self.kargs = kargs

    def __str__(self) -> str:
        interact=dict(id=0,method=self.mname,args=self.args,kargs=self.kargs,event=b"jevent(event)")
        if self.instance is not None: interact["id"]=id(self.instance)
        gen = lambda ll: (";".join(ll))+";" if ll else ""
        return f"""try{{{gen(self._befores)}interact( {stringify(interact)} );{gen(self._afters)}}} catch(e) {{_error(e,"JS")}}"""

class Binder:
    def __init__(self,btag_instance):
        self.__instance=btag_instance
    def __getattr__(self,method:str):
        m=hasattr(self.__instance,method) and getattr(self.__instance,method)
        if m and callable( m ):
            def _(*args,**kargs) -> BaseCaller:
                return BaseCaller(self.__instance,method,args,kargs)
            return _
        else:
            raise HTagException("Unknown method '%s' in '%s'"%(method,self.__instance.__class__.__name__))

    def __call__(self,callback:Callable,*args,**kargs) -> Caller:
        return Caller(self.__instance,callback,args,kargs)


class Tag(TagBase,metaclass=TagCreator): # custom tag (to inherit)
    statics: list = [] # list of "Tag", imported at start in html>head
    imports = None

    __instances__ = weakref.WeakValueDictionary()

    js: StrNonable = None  # post script, useful for js/init when tag is rendered

    @classmethod
    def find_tag(cls, obj_id:int):
        return cls.__instances__.get(obj_id, None)

    def __init__(self, *args,**kargs):
        self._hr=None
        self.parent=None    #COULD DISAPPEAR SOON
        self.event={}       # NEW !!!!
        self._callbacks_={}
        attrs={}
        selfs={}
        for k,v in kargs.items():
            if k.startswith("_"):
                attrs[k]=v
            elif k=="hr":       # provided by the HRenderer when instanciating the Tag !
                self._hr = v
            else:
                selfs[k]=v

        # own a simplified init() ?
        init=None
        if hasattr(self,"init"):
            init = getattr(self,"init")
            init = init if callable(init) else None

        if "_id" in attrs:
            raise HTagException("can't set the html attribut '_id'")
        else:
            the_id = id(self)
            attrs["_id"]=the_id   # force an @id !
            if init:
                TagBase.__init__(self, None, **attrs)
                #=================================================== in v <= 0.7.5
                # init(*args,**selfs)
                #===================================================

                #=================================================== in v > 0.7.5
                if "hr" in kargs:
                    del kargs["hr"]
                init(*args,**kargs)
                self.__dict__.update(selfs)
                #===================================================
            else:
                # if no own 'init' method, declare default args (selfs) as attributs instance
                self.__dict__.update(selfs)

                TagBase.__init__(self, *args, **attrs)
            Tag.__instances__[the_id]=self

    # new mechanism (could replace self.bind.<m>()) ... one day
    #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

    async def __on__(self,eventjs:str,*a,**ka):
        logger.info(f"callback __on__ {eventjs} {a} {ka}")
        caller = self._callbacks_[eventjs]

        for method,a,ka in [(caller.callback,a,ka)] + caller._others:
            if hasattr(method, '__self__') and method.__self__ == self:
                if asyncio.iscoroutinefunction( method ):
                    r=await method(*a,**ka)
                else:
                    r=method(*a,**ka)
            else:
                if asyncio.iscoroutinefunction( method ):
                    r=await method(self,*a,**ka)
                else:
                    r=method(self,*a,**ka)

            if isinstance(r, types.AsyncGeneratorType):
                async for i in r:
                    yield i
            elif isinstance(r, types.GeneratorType):
                for i in r:
                    yield i
            else:
                assert r is None

    def __getitem__(self,attr:str) -> Any:
        attr=attr.strip().lower()
        r=TagBase.__getitem__(self,attr)
        if r is None:
            if attr.startswith("on"):
                # self._attrs["class"]=StrClass()
                return NotBindedCaller(self,attr)

        return r

    def __setitem__(self,attr:str,value):
        if type(value) in [types.FunctionType,types.MethodType]:
            value = Caller( self, value, (), {})
            value.assignEventOnTag( self, attr )
            logger.info("Assign event '%s' (simple callback) on %s" % (attr,repr(self)))
        elif isinstance(value,Caller):
            value.assignEventOnTag( self, attr )
            logger.info("Assign event '%s' (handled) on %s" % (attr,repr(value.instance)))

        TagBase.__setitem__(self,attr,value)
    #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


    @property
    def bind(self):
        """ to bind method ! and return its js repr"""
        return Binder(self)

    #to override
    def exit(self):
        print("exit() DOES NOTHING (should be overriden)")

    def _getAllJs(self) -> list:
        """ get a list of IIFE js declared script of this tag and its children"""
        ll=[]
        if self.js:
            logger.debug("Init Script (.js) found in %s --> '%s'",repr(self),self.js)
            ll.append( self._genIIFEScript( self.js ) ) #IIFE !

        def rec(childs:Sequence):
            for i in childs:
                if isinstance(i,Tag):
                    ll.extend( i._getAllJs() )
                elif isinstance(i,TagBase):
                    rec(i._childs)

        rec(self._childs)

        return ll

    def __call__(self, js:str):
        """ Send "js to execute" (post js) now """
        if self._hr:
            current = self
        else:
            current = self.parent
            while current is not None and current.parent is not None:
                current = current.parent

        if current is None or current._hr is None:
            logger.error(f"call js is not possible, {repr(self)} is not tied to a parent/HRenderer !")
        else:
            current._hr._addInteractionScript( self._genIIFEScript(js) )

    def add(self,elt:AnyTags):
        """ override tagbase.add to feed 'tag.parent' with the parent
            TODO: make it a lot better !
        """
        if isinstance(elt,Tag):
            elt.parent = self
        elif not isinstance(elt,str) and hasattr(elt,"__iter__"):
            for i in elt:
                if isinstance(i,Tag):
                    i.parent = self
        TagBase.add(self,elt)

    def _genIIFEScript(self,js:str) -> str:
        return f"(function(tag){{ {js}\n }})(document.getElementById('{id(self)}'));"

    def _hasRender(self):
        if hasattr(self,"render"):
            render = getattr(self,"render")
            if callable(render):
                return render

    def __str__(self):
        render = self._hasRender()
        if render:
            logger.debug("Tag.__str__() : %s rendering itself with its render() method", repr(self))
            self.clear()
            render()
        else:
            logger.debug("Tag.__str__() : render str for %s", repr(self))
        return TagBase.__str__(self)


